//  icalBuddy.m
//
//  Created by Ali Rantakari on 17 June 2008
//  http://hasseg.org
//

/*
The MIT License

Copyright (c) 2008 Ali Rantakari

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/


#import <Foundation/Foundation.h>
#import <CalendarStore/CalendarStore.h>
#import <AppKit/AppKit.h>
#import <AddressBook/AddressBook.h>


#define kAppSiteURLPrefix @"http://hasseg.org/icalBuddy/"
#define kVersionCheckURL [NSURL URLWithString:@"http://hasseg.org/icalBuddy/?versioncheck=y"]


#define kDefaultTimeFormatString @"%H:%M"
#define kDefaultDateFormatString @"%Y-%m-%d"
#define kDefaultDateTimeSeparatorString @" at "


const int VERSION_MAJOR = 1;
const int VERSION_MINOR = 2;
const int VERSION_BUILD = 5;







// printOptions for calendar item printing functions
enum calItemPrintOption
{
	PRINT_OPTION_NONE = 				0,
	PRINT_OPTION_TODAY = 				(1 << 0),	// in the contex of today (for events)
	PRINT_OPTION_CALENDAR_AGNOSTIC = 	(1 << 1),	// calendar-agnostic (i.e. don't print out the calendar name)
	PRINT_OPTION_FORMAT_OUTPUT = 		(1 << 2)	// whether to use formatting in the output via ANSI escape sequences
} CalItemPrintOption;




// the string encoding to use for output
NSStringEncoding outputStrEncoding = NSUTF8StringEncoding; // default







// helper method
NSString* versionNumber()
{
	return [NSString stringWithFormat:@"%d.%d.%d", VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD];
}


// helper methods
// 		from: http://www.sveinbjorn.org/objectivec_stdout
// 		(modified to use non-deprecated version of writeToFile:...)
void NSPrint(NSString *str)
{
	[str writeToFile: @"/dev/stdout" atomically: NO encoding: outputStrEncoding error: NULL];
	//printf("%s", [str cStringUsingEncoding: outputStrEncoding]);
}

void NSPrintErr(NSString *str)
{
	[str writeToFile: @"/dev/stderr" atomically: NO encoding: outputStrEncoding error: NULL];
}





// helper method: compare three-part version number strings (e.g. "1.12.3")
NSComparisonResult versionNumberCompare(NSString *first, NSString *second)
{
	if (first != nil && second != nil)
	{
		int i;
		
		NSMutableArray *firstComponents = [NSMutableArray arrayWithCapacity:3];
		[firstComponents addObjectsFromArray:[first componentsSeparatedByString:@"."]];
		
		NSMutableArray *secondComponents = [NSMutableArray arrayWithCapacity:3];
		[secondComponents addObjectsFromArray:[second componentsSeparatedByString:@"."]];
		
		if ([firstComponents count] != [secondComponents count])
		{
			NSMutableArray *shorter;
			NSMutableArray *longer;
			if ([firstComponents count] > [secondComponents count])
			{
				shorter = secondComponents;
				longer = firstComponents;
			}
			else
			{
				shorter = firstComponents;
				longer = secondComponents;
			}
			
			NSUInteger countDiff = [longer count] - [shorter count];
			
			for (i = 0; i < countDiff; i++)
				[shorter addObject:@"0"];
		}
		
		for (i = 0; i < [firstComponents count]; i++)
		{
			int firstComponentIntVal = [[firstComponents objectAtIndex:i] intValue];
			int secondComponentIntVal = [[secondComponents objectAtIndex:i] intValue];
			if (firstComponentIntVal < secondComponentIntVal)
				return NSOrderedAscending;
			else if (firstComponentIntVal > secondComponentIntVal)
				return NSOrderedDescending;
		}
		return NSOrderedSame;
	}
	else
		return NSOrderedSame;
}




// create an NSSet from a comma-separated string,
// trimming whitespace from around each string component
NSSet* setFromCommaSeparatedStringTrimmingWhitespace(NSString *str)
{
	if (str != nil)
	{
		NSMutableSet *set = [NSMutableSet setWithCapacity:10];
		NSArray *arr = [str componentsSeparatedByString:@","];
		NSString *component;
		for (component in arr)
			[set addObject:[component stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]]];
		return set;
	}
	return [NSSet set];
}


// create an NSArray from a comma-separated string,
// trimming whitespace from around each string component
NSArray* arrayFromCommaSeparatedStringTrimmingWhitespace(NSString *str)
{
	if (str != nil)
	{
		NSMutableArray *retArr = [NSMutableArray arrayWithCapacity:10];
		NSArray *arr = [str componentsSeparatedByString:@","];
		NSString *component;
		for (component in arr)
			[retArr addObject:[component stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]]];
		return retArr;
	}
	return [NSArray array];
}




// sort function for sorting tasks:
// - sort numerically by priority except treat CalPriorityNone (0) as a special case
// - if priorities match, sort tasks that are late from their due date to be first and then
//   order alphabetically by title
NSInteger prioritySort(id task1, id task2, void *context)
{
    if ([task1 priority] < [task2 priority])
	{
		if ([task1 priority] == CalPriorityNone)
			return NSOrderedDescending;
		else
			return NSOrderedAscending;
	}
    else if ([task1 priority] > [task2 priority])
		if ([task2 priority] == CalPriorityNone)
			return NSOrderedAscending;
		else
			return NSOrderedDescending;
    else
	{
		// check if one task is late and the other is not
		NSCalendarDate *now = [NSCalendarDate calendarDate];
		BOOL task1late = NO;
		BOOL task2late = NO;
		if ([task1 dueDate] != nil)
			if ([now compare:[task1 dueDate]] == NSOrderedDescending)
				task1late = YES;
		if ([task2 dueDate] != nil)
			if ([now compare:[task2 dueDate]] == NSOrderedDescending)
				task2late = YES;
		
		if (task1late && !task2late)
			return NSOrderedAscending;
		else if (task2late && !task1late)
			return NSOrderedDescending;
		
		// neither task is, or both tasks are late -> order alphabetically by title
        return [[task1 title] compare:[task2 title]];
	}
}


// whether the two specified dates represent the same calendar day
BOOL datesRepresentSameDay(NSCalendarDate *date1, NSCalendarDate *date2)
{
	return ([date1 yearOfCommonEra] == [date2 yearOfCommonEra] && [date1 monthOfYear] == [date2 monthOfYear] && [date1 dayOfMonth] == [date2 dayOfMonth]);
}



// whether propertyName is ok to be printed, based on a set of property
// names to be included and a set of property names to be excluded
BOOL shouldPrintProperty(NSString *propertyName, NSSet *inclusionsSet, NSSet *exclusionsSet)
{
	if (propertyName == nil || (inclusionsSet == nil && exclusionsSet == nil))
		return YES;
	
	BOOL retVal = YES;
	
	if (inclusionsSet != nil)
		if (![inclusionsSet containsObject:propertyName])
			retVal = NO;
	
	if (retVal == YES && exclusionsSet != nil)
		if ([exclusionsSet containsObject:propertyName] || [exclusionsSet containsObject:@"*"])
			retVal = NO;
	
	return retVal;
}







// returns a formatted date+time
NSString* dateStr(NSDate *date, BOOL includeDate, BOOL includeTime,
					NSString *timeFormatStr, NSString *dateFormatStr,
					NSString *dateTimeSeparatorStr)
{
	if (date == nil || (!includeDate && !includeTime))
		return @"";
	
	NSCalendarDate *calDate = [date dateWithCalendarFormat:nil timeZone:nil];
	NSCalendarDate *now = [NSCalendarDate calendarDate];
	
	NSString *outputDate = nil;
	NSString *outputTime = nil;
	
	if (includeDate)
	{
		if (datesRepresentSameDay(calDate, now))
			outputDate = @"today";
		else if (datesRepresentSameDay(calDate, [now dateByAddingYears:0 months:0 days:1 hours:0 minutes:0 seconds:0]))
			outputDate = @"tomorrow";
		else
		{
			NSString *useDateFormatStr = (dateFormatStr != nil) ? dateFormatStr : kDefaultDateFormatString;
			outputDate = [date descriptionWithCalendarFormat:useDateFormatStr timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]];
		}
	}
	
	if (includeTime)
	{
		NSString *useTimeFormatStr = (timeFormatStr != nil) ? timeFormatStr : kDefaultTimeFormatString;
		outputTime = [date descriptionWithCalendarFormat:useTimeFormatStr timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]];
	}
	
	if (outputDate != nil && outputTime == nil)
		return outputDate;
	else if (outputDate == nil && outputTime != nil)
		return outputTime;
	else
	{
		NSString *dateTimeSeparatorToUse = (dateTimeSeparatorStr != nil) ? dateTimeSeparatorStr : kDefaultDateTimeSeparatorString;
		return [NSString stringWithFormat:@"%@%@%@", outputDate, dateTimeSeparatorToUse, outputTime];
	}
}










// pretty-prints out the specified event
void printCalEvent(CalEvent *event, int printOptions,
					NSString *timeFormatStr, NSString *dateFormatStr, NSString *dateTimeSeparatorStr,
					NSSet *includedProps, NSSet *excludedProps)
{
	if (event != nil)
	{
		NSString *bullet = @"*";
		BOOL formatOutput = (printOptions & PRINT_OPTION_FORMAT_OUTPUT);
		
		if ([[[event calendar] type] isEqualToString:CalCalendarTypeBirthday])
		{
			// special case for events in the Birthdays calendar (they don't seem to have titles
			// so we have to use the URI to find the ABPerson from the Address Book
			// and print their name from there)
			
			NSString *personId = [[NSString stringWithFormat:@"%@", [event url]] stringByReplacingOccurrencesOfString:@"addressbook://" withString:@""];
			ABRecord *person = [[ABAddressBook sharedAddressBook] recordForUniqueId:personId];
			
			if (person != nil)
			{
				if ([person isMemberOfClass: [ABPerson class]])
				{
					NSString *thisTitle = [NSString stringWithFormat:@"%@ %@'s Birthday", [person valueForProperty:kABFirstNameProperty], [person valueForProperty:kABLastNameProperty]];
					if (printOptions & PRINT_OPTION_CALENDAR_AGNOSTIC)
						NSPrint([NSString stringWithFormat: @"%@%@ %@%@\n", formatOutput?@"\033[01m":@"", bullet, thisTitle, formatOutput?@"\033[0m":@""]);
					else
						NSPrint([NSString stringWithFormat: @"%@%@ %@ (%@)%@\n", formatOutput?@"\033[01m":@"", bullet, thisTitle, [[event calendar] title], formatOutput?@"\033[0m":@""]);
					
					if (!(printOptions & PRINT_OPTION_TODAY))
						NSPrint([NSString stringWithFormat: @"    %@\n",
								dateStr([event startDate], true, false, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
								]);
				}
			}
		}
		else
		{
			NSCalendarDate *now = [NSCalendarDate calendarDate];
			
			// title
			if (printOptions & PRINT_OPTION_CALENDAR_AGNOSTIC)
				NSPrint([NSString stringWithFormat: @"%@%@ %@%@\n", formatOutput?@"\033[01m":@"", bullet, [event title], formatOutput?@"\033[0m":@""]);
			else
				NSPrint([NSString stringWithFormat: @"%@%@ %@ (%@)%@\n", formatOutput?@"\033[01m":@"", bullet, [event title], [[event calendar] title], formatOutput?@"\033[0m":@""]);
			
			// location
			if (shouldPrintProperty(@"location", includedProps, excludedProps))
				if ([event location] != nil && ![[[event location] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] isEqualToString:@""])
					NSPrint([NSString stringWithFormat: @"    location: %@\n", [event location]]);
			
			// notes
			if (shouldPrintProperty(@"notes", includedProps, excludedProps))
				if ([event notes] != nil && ![[[event notes] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] isEqualToString:@""])
					NSPrint([NSString stringWithFormat: @"    notes: %@\n", [[event notes] stringByReplacingOccurrencesOfString:@"\n" withString:@"\n           "]]);
			
			// url
			if (shouldPrintProperty(@"url", includedProps, excludedProps))
				if ([event url] != nil)
					NSPrint([NSString stringWithFormat: @"    url: %@\n", [event url]]);
			
			// start/end dates & times
			if (shouldPrintProperty(@"startDate", includedProps, excludedProps) &&
				shouldPrintProperty(@"endDate", includedProps, excludedProps))
			{
				if ( !(printOptions & PRINT_OPTION_TODAY) || ( (printOptions & PRINT_OPTION_TODAY) && ![event isAllDay] ) )
				{
					if ([[event startDate] isEqualToDate:[event endDate]])
					{
						if (printOptions & PRINT_OPTION_TODAY)
							NSPrint([NSString stringWithFormat: @"    %@\n",
									dateStr([event startDate], false, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
									]);
						else
							NSPrint([NSString stringWithFormat: @"    %@\n",
									dateStr([event startDate], true, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
									]);
					}
					else
					{
						if (printOptions & PRINT_OPTION_TODAY)
						{
							BOOL startsToday = datesRepresentSameDay(now, [[event startDate] dateWithCalendarFormat:nil timeZone:nil]);
							BOOL endsToday = datesRepresentSameDay(now, [[event endDate] dateWithCalendarFormat:nil timeZone:nil]);
							if (startsToday && endsToday)
								NSPrint([NSString stringWithFormat: @"    %@ - %@\n",
										dateStr([event startDate], false, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr),
										dateStr([event endDate], false, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
										]);
							else if (startsToday)
								NSPrint([NSString stringWithFormat: @"    %@ - ...\n",
										dateStr([event startDate], false, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
										]);
							else if (endsToday)
								NSPrint([NSString stringWithFormat: @"    ... - %@\n",
										dateStr([event endDate], false, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
										]);
						}
						else
						{
							if ([event isAllDay])
							{
								NSInteger daysDiff;
								[[[event endDate] dateWithCalendarFormat:nil timeZone:nil]
									years:NULL months:NULL days:&daysDiff hours:NULL minutes:NULL seconds:NULL
									sinceDate:[[event startDate] dateWithCalendarFormat:nil timeZone:nil]
								];
								if (daysDiff > 1)
								{
									// all-day events technically span from <start day> at 00:00 to <end day+1> at 00:00 even though
									// we want them displayed as only spanning from <start day> to <end day>
									NSCalendarDate *endDateMinusOneDay = [[[event endDate] dateWithCalendarFormat:nil timeZone:nil] dateByAddingYears:0 months:0 days:-1 hours:0 minutes:0 seconds:0];
									NSPrint([NSString stringWithFormat: @"    %@ - %@\n",
											dateStr([event startDate], true, false, timeFormatStr, dateFormatStr, dateTimeSeparatorStr),
											dateStr(endDateMinusOneDay, true, false, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
											]);
								}
								else
								{
									NSPrint([NSString stringWithFormat: @"    %@\n",
											dateStr([event startDate], true, false, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
											]);
								}
							}
							else
							{
								NSString *startDateFormattedStr = dateStr([event startDate], true, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr);
								NSString *endDateFormattedStr;
								if (datesRepresentSameDay([[event startDate] dateWithCalendarFormat:nil timeZone:nil], [[event endDate] dateWithCalendarFormat:nil timeZone:nil]))
									endDateFormattedStr = dateStr([event endDate], false, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr);
								else
									endDateFormattedStr = dateStr([event endDate], true, true, timeFormatStr, dateFormatStr, dateTimeSeparatorStr);
								NSPrint([NSString stringWithFormat: @"    %@ - %@\n", startDateFormattedStr, endDateFormattedStr]);
							}
						}
					}
				}
			}
		}
	}
}






// pretty-prints out the specified task
void printCalTask(CalTask *task, int printOptions,
					NSString *timeFormatStr, NSString *dateFormatStr, NSString *dateTimeSeparatorStr,
					NSSet *includedProps, NSSet *excludedProps)
{
	if (task != nil)
	{
		NSCalendarDate *now = [NSCalendarDate calendarDate];
		NSString *bullet = @"*";
		BOOL formatOutput = (printOptions & PRINT_OPTION_FORMAT_OUTPUT);
		
		// if task's due date is in the past, use ! in place of the bullet point (*)
		if ([task dueDate] != nil)
			if ([now compare:[task dueDate]] == NSOrderedDescending)
				bullet = @"!";
		
		// title
		if (printOptions & PRINT_OPTION_CALENDAR_AGNOSTIC)
			NSPrint([NSString stringWithFormat: @"%@%@ %@%@\n", formatOutput?@"\033[01m":@"", bullet, [task title], formatOutput?@"\033[0m":@""]);
		else
			NSPrint([NSString stringWithFormat: @"%@%@ %@ (%@)%@\n", formatOutput?@"\033[01m":@"", bullet, [task title], [[task calendar] title], formatOutput?@"\033[0m":@""]);
		
		// notes
		if (shouldPrintProperty(@"notes", includedProps, excludedProps))
			if ([task notes] != nil && ![[[task notes] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] isEqualToString:@""])
				NSPrint([NSString stringWithFormat: @"    notes: %@\n", [[task notes] stringByReplacingOccurrencesOfString:@"\n" withString:@"\n           "]]);
		
		// url
		if (shouldPrintProperty(@"url", includedProps, excludedProps))
			if ([task url] != nil)
				NSPrint([NSString stringWithFormat: @"    url: %@\n", [task url]]);
		
		// due date
		if (shouldPrintProperty(@"dueDate", includedProps, excludedProps))
		{
			if ([task dueDate] != nil)
			{
				NSPrint([NSString stringWithFormat: @"    due: %@\n",
						dateStr([task dueDate], true, false, timeFormatStr, dateFormatStr, dateTimeSeparatorStr)
						]);
			}
		}
		
		// priority
		if (shouldPrintProperty(@"priority", includedProps, excludedProps))
		{
			if ([task priority] != CalPriorityNone)
			{
				switch([task priority])
				{
					case CalPriorityHigh:
						NSPrint(@"    priority: high\n");
						break;
					case CalPriorityMedium:
						NSPrint(@"    priority: medium\n");
						break;
					case CalPriorityLow:
						NSPrint(@"    priority: low\n");
						break;
					default:
						NSPrint([NSString stringWithFormat: @"    priority: %d\n", [task priority]]);
						break;
				}
			}
		}
	}
}










int main(int argc, char *argv[])
{
	NSAutoreleasePool *autoReleasePool = [[NSAutoreleasePool alloc] init];
	
	
	
	// variables for arguments
	NSString *arg_output = nil;
	BOOL arg_separateByCalendar = NO;
	BOOL arg_separateByDate = NO;
	NSArray *arg_includeCals = nil;
	NSArray *arg_excludeCals = nil;
	BOOL arg_updatesCheck = NO;
	BOOL arg_printVersion = NO;
	NSString *arg_timeFormatStr = nil;
	NSString *arg_dateFormatStr = nil;
	NSString *arg_dateTimeSeparatorStr = nil;
	NSSet *arg_includedEventProperties = nil;
	NSSet *arg_excludedEventProperties = nil;
	NSSet *arg_includedTaskProperties = nil;
	NSSet *arg_excludedTaskProperties = nil;
	BOOL arg_includeOnlyEventsFromNowOn = NO;
	BOOL arg_useFormatting = NO;
	BOOL arg_noCalendarNames = NO;
	NSString *arg_strEncoding = nil;
	
	// get arguments
	
	if (argc > 1)
		arg_output = [NSString stringWithCString: argv[argc-1] encoding: NSASCIIStringEncoding];
	
	int i;
	for (i = 0; i < argc; i++)
	{
		if ((strcmp(argv[i], "-sc") == 0) || (strcmp(argv[i], "--separateByCalendar") == 0))
			arg_separateByCalendar = YES;
		else if ((strcmp(argv[i], "-sd") == 0) || (strcmp(argv[i], "--separateByDate") == 0))
			arg_separateByDate = YES;
		else if ((strcmp(argv[i], "-u") == 0) || (strcmp(argv[i], "--checkForUpdates") == 0))
			arg_updatesCheck = YES;
		else if ((strcmp(argv[i], "-V") == 0) || (strcmp(argv[i], "--version") == 0))
			arg_printVersion = YES;
		else if ((strcmp(argv[i], "-n") == 0) || (strcmp(argv[i], "--includeOnlyEventsFromNowOn") == 0))
			arg_includeOnlyEventsFromNowOn = YES;
		else if ((strcmp(argv[i], "-f") == 0) || (strcmp(argv[i], "--formatOutput") == 0))
			arg_useFormatting = YES;
		else if ((strcmp(argv[i], "-nc") == 0) || (strcmp(argv[i], "--noCalendarNames") == 0))
			arg_noCalendarNames = YES;
	}
	
	NSString *includeCalsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"ic"];
	if (includeCalsStr == nil)
		includeCalsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-includeCals"];
	if (includeCalsStr != nil)
		arg_includeCals = arrayFromCommaSeparatedStringTrimmingWhitespace(includeCalsStr);
	
	NSString *excludeCalsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"ec"];
	if (excludeCalsStr == nil)
		excludeCalsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-excludeCals"];
	if (excludeCalsStr != nil)
		arg_excludeCals = arrayFromCommaSeparatedStringTrimmingWhitespace(excludeCalsStr);
	
	arg_timeFormatStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"tf"];
	if (arg_timeFormatStr == nil)
		arg_timeFormatStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-timeFormat"];
	
	arg_dateFormatStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"df"];
	if (arg_dateFormatStr == nil)
		arg_dateFormatStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-dateFormat"];
	
	arg_dateTimeSeparatorStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"dts"];
	if (arg_dateTimeSeparatorStr == nil)
		arg_dateTimeSeparatorStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-dateTimeSeparator"];
	
	NSString *includeEventPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"iep"];
	if (includeEventPropsStr == nil)
		includeEventPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-includeEventProps"];
	if (includeEventPropsStr != nil)
		arg_includedEventProperties = setFromCommaSeparatedStringTrimmingWhitespace(includeEventPropsStr);
	
	NSString *excludeEventPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"eep"];
	if (excludeEventPropsStr == nil)
		excludeEventPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-excludeEventProps"];
	if (excludeEventPropsStr != nil)
		arg_excludedEventProperties = setFromCommaSeparatedStringTrimmingWhitespace(excludeEventPropsStr);
	
	NSString *includeTaskPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"itp"];
	if (includeTaskPropsStr == nil)
		includeTaskPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-includeTaskProps"];
	if (includeTaskPropsStr != nil)
		arg_includedTaskProperties = setFromCommaSeparatedStringTrimmingWhitespace(includeTaskPropsStr);
	
	NSString *excludeTaskPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"etp"];
	if (excludeTaskPropsStr == nil)
		excludeTaskPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-excludeTaskProps"];
	if (excludeTaskPropsStr != nil)
		arg_excludedTaskProperties = setFromCommaSeparatedStringTrimmingWhitespace(excludeTaskPropsStr);
	
	arg_strEncoding = [[NSUserDefaults standardUserDefaults] stringForKey:@"-strEncoding"];
	if (arg_strEncoding != nil)
	{
		// process provided output string encoding argument
		arg_strEncoding = [arg_strEncoding stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
		NSStringEncoding matchedEncoding = 0;
		const NSStringEncoding *availableEncoding = [NSString availableStringEncodings];
		while(*availableEncoding != 0)
		{
			if ([[NSString localizedNameOfStringEncoding: *availableEncoding] isEqualToString:arg_strEncoding])
			{
				matchedEncoding = *availableEncoding;
				break;
			}
			availableEncoding++;
		}
		if (matchedEncoding != 0)
			outputStrEncoding = matchedEncoding;
		else
			NSPrintErr([NSString stringWithFormat:@"\nError: Invalid string encoding argument: \"%@\". Run \"icalBuddy strEncodings\" to see all the possible values. Using default encoding \"%@\".\n\n", arg_strEncoding, [NSString localizedNameOfStringEncoding: outputStrEncoding]]);
	}
	
	
	
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print version and exit
	// ------------------------------------------------------------------
	if (arg_printVersion)
	{
		NSPrint([NSString stringWithFormat:@"%@\n", versionNumber()]);
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// check for updates
	// ------------------------------------------------------------------
	else if (arg_updatesCheck)
	{
		NSPrint(@"Checking for updates... ");
		
		NSURL *url = kVersionCheckURL;
		NSURLRequest *request = [NSURLRequest requestWithURL:url cachePolicy:NSURLRequestReloadIgnoringCacheData timeoutInterval:10.0];
		
		NSHTTPURLResponse *response;
		NSError *error;
		[NSURLConnection sendSynchronousRequest:request returningResponse:&response error:&error];
		
		if (error == nil && response != nil)
		{
			NSInteger statusCode = [response statusCode];
			if (statusCode >= 400)
			{
				NSPrint([NSString stringWithFormat:@"...HTTP connection failed.\n\nStatus code %d: \"%@\"\n\n",
									statusCode, [NSHTTPURLResponse localizedStringForStatusCode:statusCode]]);
			}
			else
			{
				NSString *latestVersionString = [[response allHeaderFields] valueForKey:@"Orghassegsoftwarelatestversion"];
				NSString *currentVersionString = versionNumber();
				
				if (latestVersionString == nil)
					NSPrintErr([NSString stringWithFormat:@"...failed.\n\nError reading latest version number from HTTP header field.\n\n"]);
				else
				{
					if (versionNumberCompare(currentVersionString, latestVersionString) == NSOrderedAscending)
					{
						NSPrint([NSString stringWithFormat:@"...update found! (latest: %@  current: %@)\n\n", latestVersionString, currentVersionString]);
						NSPrint([NSString stringWithFormat:@"Navigate to the following URL to see the release notes and download the latest version:\n\n%@?currentversion=%@\n\n", kAppSiteURLPrefix, currentVersionString]);
						char inputChar;
						while(inputChar != 'y' && inputChar != 'n' && inputChar != 'Y' && inputChar != 'N' && inputChar != '\n')
						{
							NSPrint(@"Do you want to navigate to this URL now? [y/n] ");
							scanf("%s&*c",&inputChar);
						}
						
						if (inputChar == 'y' || inputChar == 'Y')
							[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:[NSString stringWithFormat: @"%@?currentversion=%@", kAppSiteURLPrefix, currentVersionString]]];
					}
					else
						NSPrint([NSString stringWithFormat:@"...you're up to date! (latest: %@  current: %@)\n\n", latestVersionString, currentVersionString]);
				}
			}
		}
		else
		{
			NSPrintErr([NSString stringWithFormat:@"...connection failed.\n\nError: - %@ %@\n\n",
				[error localizedDescription],
				[[error userInfo] objectForKey:NSErrorFailingURLStringKey]]);
		}
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print possible values for the string encoding argument and exit
	// ------------------------------------------------------------------
	else if ([arg_output isEqualToString:@"strEncodings"])
	{
		NSPrint(@"\nAvailable String encodings (you can use one of these\nas an argument to the --strEncoding option):\n\n");
		const NSStringEncoding *availableEncoding = [NSString availableStringEncodings];
		while(*availableEncoding != 0)
		{
			NSPrint([NSString stringWithFormat:@"%@\n", [NSString localizedNameOfStringEncoding: *availableEncoding]]);
			availableEncoding++;
		}
		NSPrint(@"\n");
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print all calendars
	// ------------------------------------------------------------------
	else if ([arg_output isEqualToString:@"calendars"])
	{
		NSArray *allCalendars = [[CalCalendarStore defaultCalendarStore] calendars];
		
		CalCalendar *cal;
		for (cal in allCalendars)
		{
			NSPrint([NSString stringWithFormat: @"* %@\n  uid: %@\n", [cal title], [cal uid]]);
		}
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print EVENTS
	// ------------------------------------------------------------------
	else if ([arg_output hasPrefix:@"eventsToday"] || [arg_output isEqualToString:@"eventsNow"])
	{
		// get all calendars
		NSMutableArray *allCalendars = [NSMutableArray arrayWithCapacity:10];
		[allCalendars addObjectsFromArray: [[CalCalendarStore defaultCalendarStore] calendars]];
		
		// filter calendars to use:
		// - first include what has been specified to be included,
		// - then exclude what has been specified to be excluded
		if (arg_includeCals != nil)
			[allCalendars filterUsingPredicate: [NSPredicate predicateWithFormat:@"(uid IN %@) OR (title IN %@)", arg_includeCals, arg_includeCals]];
		if (arg_excludeCals != nil)
			[allCalendars filterUsingPredicate: [NSPredicate predicateWithFormat:@"(NOT(uid IN %@)) AND (NOT(title IN %@))", arg_excludeCals, arg_excludeCals]];
		
		// default print options
		int events_printOptions = 
			PRINT_OPTION_TODAY | 
			(arg_useFormatting ? PRINT_OPTION_FORMAT_OUTPUT : PRINT_OPTION_NONE) |
			(arg_noCalendarNames ? PRINT_OPTION_CALENDAR_AGNOSTIC : PRINT_OPTION_NONE);
		
		// get start and end dates for predicate
		NSCalendarDate *now = [NSCalendarDate calendarDate];
		NSCalendarDate *eventsDateRangeStart = nil;
		NSCalendarDate *eventsDateRangeEnd = nil;
		if ([arg_output hasPrefix:@"eventsToday"])
		{
			eventsDateRangeStart = [NSCalendarDate dateWithYear:[now yearOfCommonEra] month:[now monthOfYear] day:[now dayOfMonth] hour:0 minute:0 second:0 timeZone:[now timeZone]];
			eventsDateRangeEnd = [NSCalendarDate dateWithYear:[now yearOfCommonEra] month:[now monthOfYear] day:[now dayOfMonth] hour:23 minute:59 second:59 timeZone:[now timeZone]];
		}
		else if ([arg_output isEqualToString:@"eventsNow"])
		{
			eventsDateRangeStart = now;
			eventsDateRangeEnd = now;
		}
		NSCAssert((eventsDateRangeStart != nil && eventsDateRangeEnd != nil), @"start or end date is nil");
		
		// expand end date if NUM in "eventsToday+NUM" is specified
		if ([arg_output hasPrefix:@"eventsToday"])
		{
			NSRange arg_output_plusSymbolRange = [arg_output rangeOfString:@"+"];
			if (arg_output_plusSymbolRange.location != NSNotFound)
			{
				NSInteger daysToAddToRange = [[arg_output substringFromIndex:(arg_output_plusSymbolRange.location+arg_output_plusSymbolRange.length)] intValue];
				eventsDateRangeEnd = [eventsDateRangeEnd dateByAddingYears:0 months:0 days:daysToAddToRange hours:0 minutes:0 seconds:0];
				events_printOptions -= PRINT_OPTION_TODAY;
			}
		}
		
		// make predicate for getting all events for today + use it to get the events
		NSPredicate *eventsPredicate = [CalCalendarStore eventPredicateWithStartDate:((arg_includeOnlyEventsFromNowOn)?now:eventsDateRangeStart) endDate:eventsDateRangeEnd calendars:allCalendars];
		NSArray *eventsArr = [[CalCalendarStore defaultCalendarStore] eventsWithPredicate:eventsPredicate];
		
		// print the events out
		if (arg_separateByCalendar)
		{
			// separated by calendar
			BOOL titlePrintedForCurrentCalendar;
			BOOL currentIsFirstPrintedCalendar = YES;
			CalCalendar *cal;
			for (cal in allCalendars)
			{
				titlePrintedForCurrentCalendar = NO;
				CalEvent *event;
				for (event in eventsArr)
				{
					if ([[cal uid] isEqualToString: [[event calendar] uid]])
					{
						if (!titlePrintedForCurrentCalendar)
						{
							if (!currentIsFirstPrintedCalendar)
								NSPrint(@"\n");
							NSPrint([NSString stringWithFormat: @"%@:\n------------------------\n", [cal title]]);
							titlePrintedForCurrentCalendar = YES;
							currentIsFirstPrintedCalendar = NO;
						}
						
						printCalEvent(event, ( events_printOptions | PRINT_OPTION_CALENDAR_AGNOSTIC ),
										arg_timeFormatStr, arg_dateFormatStr, arg_dateTimeSeparatorStr,
										arg_includedEventProperties, arg_excludedEventProperties);
					}
				}
			}
		}
		else
		{
			// not separated by calendar (calendar specified at the end of event titles)
			CalEvent *event;
			for (event in eventsArr)
			{
				printCalEvent(event, events_printOptions,
								arg_timeFormatStr, arg_dateFormatStr, arg_dateTimeSeparatorStr,
								arg_includedEventProperties, arg_excludedEventProperties);
			}
		}
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print all uncompleted TASKS
	// ------------------------------------------------------------------
	else if ([arg_output isEqualToString:@"uncompletedTasks"])
	{
		// get all calendars
		NSMutableArray *allCalendars = [NSMutableArray arrayWithCapacity:10];
		[allCalendars addObjectsFromArray: [[CalCalendarStore defaultCalendarStore] calendars]];
		
		// filter calendars to use:
		// - first include what has been specified to be included,
		// - then exclude what has been specified to be excluded
		if (arg_includeCals != nil)
			[allCalendars filterUsingPredicate: [NSPredicate predicateWithFormat:@"(uid IN %@) OR (title IN %@)", arg_includeCals, arg_includeCals]];
		if (arg_excludeCals != nil)
			[allCalendars filterUsingPredicate: [NSPredicate predicateWithFormat:@"(NOT(uid IN %@)) AND (NOT(title IN %@))", arg_excludeCals, arg_excludeCals]];
		
		// make predicate for uncompleted tasks in all calendars and use it to get the tasks
		NSPredicate *uncompletedTasksPredicate = [CalCalendarStore taskPredicateWithUncompletedTasks:allCalendars];
		NSArray *uncompletedTasks = [[CalCalendarStore defaultCalendarStore] tasksWithPredicate:uncompletedTasksPredicate];
		
		// sort the tasks by priority
		uncompletedTasks = [uncompletedTasks sortedArrayUsingFunction:prioritySort context:NULL];
		
		// default print options
		int tasks_printOptions = 
			(arg_useFormatting ? PRINT_OPTION_FORMAT_OUTPUT : PRINT_OPTION_NONE) |
			(arg_noCalendarNames ? PRINT_OPTION_CALENDAR_AGNOSTIC : PRINT_OPTION_NONE);
		
		
		// print out the tasks
		if (arg_separateByCalendar)
		{
			// separated by calendar
			BOOL titlePrintedForCurrentCalendar;
			BOOL currentIsFirstPrintedCalendar = YES;
			CalCalendar *cal;
			for (cal in allCalendars)
			{
				titlePrintedForCurrentCalendar = NO;
				CalTask *task;
				for (task in uncompletedTasks)
				{
					if ([[cal uid] isEqualToString: [[task calendar] uid]])
					{
						if (!titlePrintedForCurrentCalendar)
						{
							if (!currentIsFirstPrintedCalendar)
								NSPrint(@"\n");
							NSPrint([NSString stringWithFormat: @"%@:\n------------------------\n", [cal title]]);
							titlePrintedForCurrentCalendar = YES;
							currentIsFirstPrintedCalendar = NO;
						}
						printCalTask(task, (tasks_printOptions | PRINT_OPTION_CALENDAR_AGNOSTIC),
										arg_timeFormatStr, arg_dateFormatStr, arg_dateTimeSeparatorStr,
										arg_includedTaskProperties, arg_excludedTaskProperties);
					}
				}
			}
		}
		else
		{
			// not separated by calendar (calendar specified at the end of the title)
			CalTask *task;
			for (task in uncompletedTasks)
			{
				printCalTask(task, tasks_printOptions,
								arg_timeFormatStr, arg_dateFormatStr, arg_dateTimeSeparatorStr,
								arg_includedTaskProperties, arg_excludedTaskProperties);
			}
		}
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	else
	{
		NSPrint(@"\n");
		NSPrint([NSString stringWithFormat: @"USAGE: %s <options> <output_type>\n", argv[0]]);
		NSPrint(@"\n");
		NSPrint(@"<output_type> specifies what to output.\n");
		NSPrint(@"Possible values for it are:\n");
		NSPrint(@"  'eventsToday'      (events occurring today)\n");
		NSPrint(@"  'eventsToday+NUM'  (events occurring between today\n");
		NSPrint(@"                      and NUM days into the future (where\n");
		NSPrint(@"                      NUM is an integer value))\n");
		NSPrint(@"  'eventsNow'        (events occurring at present time)\n");
		NSPrint(@"  'uncompletedTasks' (uncompleted tasks)\n");
		NSPrint(@"  'calendars'        (all calendars)\n");
		NSPrint(@"  'strEncodings'     (all the possible string encodings)\n");
		NSPrint(@"\n");
		NSPrint(@"Options:\n");
		NSPrint(@"\n");
		NSPrint(@"  -V    (or: --version) Print version number and exit.\n");
		NSPrint(@"\n");
		NSPrint(@"  -u    (or: --checkForUpdates) Check for updates.\n");
		NSPrint(@"\n");
		NSPrint(@"  -sc   (or: --separateByCalendar) Separate events/tasks\n");
		NSPrint(@"        by calendar.\n");
		NSPrint(@"\n");
		NSPrint(@"  -f    (or: --formatOutput)\n");
		NSPrint(@"        Format the output with ANSI escape sequences.\n");
		NSPrint(@"\n");
		NSPrint(@"  -nc   (or: --noCalendarNames) Omit calendar names from the\n");
		NSPrint(@"        output (does not apply if the -sc argument is used.)\n");
		NSPrint(@"\n");
		NSPrint(@"  --strEncoding\n");
		NSPrint(@"        Use the specified string encoding for the output.\n");
		NSPrint(@"        Run the app once with the 'strEncodings' output_type\n");
		NSPrint(@"        value to see all the possible values you can use here.\n");
		NSPrint(@"\n");
		NSPrint(@"  -tf   (or: --timeFormat)\n");
		NSPrint(@"        Specify the formatting of times in the output.\n");
		NSPrint(@"        See http://tinyurl.com/b9mgyp [apple.com] for\n");
		NSPrint(@"        documentation on the syntax to use.\n");
		NSPrint(@"\n");
		NSPrint(@"  -df   (or: --dateFormat)\n");
		NSPrint(@"        Specify the formatting of dates in the output.\n");
		NSPrint(@"        See http://tinyurl.com/b9mgyp [apple.com] for\n");
		NSPrint(@"        documentation on the syntax to use.\n");
		NSPrint(@"\n");
		NSPrint(@"  -dts  (or: --dateTimeSeparator)\n");
		NSPrint(@"        Specify the separator string to use between\n");
		NSPrint(@"        dates and times in the output. The default is\n");
		NSPrint(@"        \" at \".\n");
		NSPrint(@"\n");
		NSPrint(@"  -n    (or: --includeOnlyEventsFromNowOn)\n");
		NSPrint(@"        If the output_type value 'eventsToday' is\n");
		NSPrint(@"        used, only output events that occur between the\n");
		NSPrint(@"        current date and the end of the day (as opposed to\n");
		NSPrint(@"        events that occur between the start and end of the\n");
		NSPrint(@"        day, as by default.)\n");
		NSPrint(@"\n");
		NSPrint(@"  -ic   (or: --includeCals) Which calendars to include.\n");
		NSPrint(@"        Must be a comma-separated list of calendar UIDs (which\n");
		NSPrint(@"        you can find out by using the 'calendars' output parameter.)\n");
		NSPrint(@"        You can also use titles, but remember that they may not be\n");
		NSPrint(@"        unique (as opposed to UIDs.)\n");
		NSPrint(@"        The -i and -e parameters will be handled in the order of\n");
		NSPrint(@"        first include, then exclude.\n");
		NSPrint(@"\n");
		NSPrint(@"  -ec   (or: --excludeCals) Which calendars to exclude.\n");
		NSPrint(@"        Must be a comma-separated list of calendar UIDs (which\n");
		NSPrint(@"        you can find out by using the 'calendars' output parameter.)\n");
		NSPrint(@"        You can also use titles, but remember that they may not be\n");
		NSPrint(@"        unique (as opposed to UIDs.)\n");
		NSPrint(@"        The -i and -e parameters will be handled in the order of\n");
		NSPrint(@"        first include, then exclude.\n");
		NSPrint(@"\n");
		NSPrint(@"  -iep  (or: --includeEventProps)\n");
		NSPrint(@"        Which event properties to include in the output. Must be\n");
		NSPrint(@"        a comma-separated list property names. Possible\n");
		NSPrint(@"        property names are: location, url, notes, startDate\n");
		NSPrint(@"        and endDate.\n");
		NSPrint(@"\n");
		NSPrint(@"  -eep  (or: --excludeEventProps)\n");
		NSPrint(@"        Which event properties to exclude from the output. Must be\n");
		NSPrint(@"        a comma-separated list property names. A value of \"*\"\n");
		NSPrint(@"        will exclude all properties and make sure just the title\n");
		NSPrint(@"        is printed.\n");
		NSPrint(@"\n");
		NSPrint(@"  -itp  (or: --includeTaskProps)\n");
		NSPrint(@"        Which task properties to include in the output. Must be\n");
		NSPrint(@"        a comma-separated list property names. Possible\n");
		NSPrint(@"        property names are: url, notes, dueDate and priority.\n");
		NSPrint(@"\n");
		NSPrint(@"  -etp  (or: --excludeTaskProps)\n");
		NSPrint(@"        Which task properties to exclude from the output. Must be\n");
		NSPrint(@"        a comma-separated list property names. A value of \"*\"\n");
		NSPrint(@"        will exclude all properties and make sure just the title\n");
		NSPrint(@"        is printed.\n");
		NSPrint(@"\n");
		NSPrint([NSString stringWithFormat: @"Version %@\n", versionNumber()]);
		NSPrint(@"(c) 2008-2009 Ali Rantakari, http://hasseg.org\n");
		NSPrint(@"\n");
	}
	
	
	[autoReleasePool release];
	return(0);
}